from retaggr.boorus.base import Booru

# External imports
import asyncio
import functools
import requests as fuck_aiohttp
from fake_useragent import UserAgent
from lxml import html

class Iqdb(Booru):
    """Reverse searches https://iqdb.org for images.

    This booru does not required images to be downloaded before searching.

    This method may have unexpected failures relating to HTML parsing. This is because IQDB does not officially offer an API and the raw HTML
    parsing is not functional if the page is not fully loaded.

    :param min_score: Minimum search match percentage needed.
    :type min_score: float
    """
    host = "https://iqdb.org"
    download_required = False

    def __init__(self, min_score):
        self.min_score = min_score
        self.ua = UserAgent()

    async def search_image(self, url: str):
        """Reverse search the Booru for ``url``.
        """
        results = []

        params = {"url" : url}
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.post, self.host, headers={"User-Agent": self.ua.random}, params=params))

        doc = html.fromstring(r.text)
        tables = doc.xpath("//div[@id='pages']/div/table/tr/td")

        row = 6
        while row < len(tables):
            # Percent similair
            if (tables[row].text) is None:
                row = row + 6
                continue
            try:
                percent = float(str.split(tables[row].text)[0][:-1])
                # Create tags list
                tags_str = tables[row-2].xpath("//a/img")
                temp_tags = tags_str[0].get('alt').split("Tags: ", 1)[1]
                tags = [x.lower().replace(',', '') for x in temp_tags.split()] # This is because IQDB searches zerochan, a site that doesn't sanitize it's tags (adding capitalization and commas which are illegal in tag names.)
                if percent > self.min_score:
                    results.extend(tags)
            except:
                pass
            row = row + 6

        return results

    async def search_tag(self, tag: str):
        """Reverse search the booru for tag data.
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")


class Booru:
    """Base class for a booru.
    
    All the booru classes must derive from this class.

    :ivar host: The base URL for the reverse image domain. This is not an API endpoint, but can be a link to IQDB or something similar.
    :vartype host: str
    :ivar download_required: Determines if the ``search_image()`` function will download the image to be searched beforehand or not.
    :vartype download_required: bool"""
    host = None
    download_required = False

    def __init__(self):
        raise NotImplementedError("Expand this method to include all needed keys.")

    async def search_image(self, url: str):
        """Reverse search the Booru for ``url``.
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")

    async def search_tag(self, tag: str):
        """Reverse search the booru for tag data.
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")



from retaggr.boorus.base import Booru

# External imports
import hashlib
from aiohttp_requests import requests
import xml.etree.ElementTree as ET

class Paheal(Booru):
    """Reverse searches https://rule34.paheal.net for images.

    This booru does require images to be downloaded before searching.
    """
    host = None
    download_required = True

    def __init__(self):
        pass

    async def search_image(self, url: str):
        """Reverse search the Booru for ``url``.
        """
        m = hashlib.md5()
        r = await requests.get(url)
        async for data in r.content.iter_chunked(8192):
            m.update(data)
        md5_hash = m.hexdigest()
        paheal_request = await requests.get(f"http://rule34.paheal.net/api/danbooru/find_posts?md5={md5_hash}")
        xml_tree = ET.fromstring(await paheal_request.text())

        tags = [] # pointless but just in case it's zero, it won't error out.
        for post in xml_tree:
            for tag in post.attrib["tags"].split(): 
                tags.append(tag.lower())
        return tags

    async def search_tag(self, tag: str):
        """Reverse search the booru for tag data.
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")



from retaggr.config import ReverseSearchConfig

# Boorus
from retaggr.boorus.danbooru import Danbooru
from retaggr.boorus.e621 import E621
from retaggr.boorus.iqdb import Iqdb
from retaggr.boorus.paheal import Paheal

# Exceptions
from retaggr.errors import MissingAPIKeysException, NotAValidBooruException

class ReverseSearch:
    r"""Core class used for Reverse Searching. 
    
    This class can only be instantiated with a :class:`ReverseSearchConfig` instance.

    All listed methods can only be ran from an asynchronous context.

    :ivar accessible_boorus: The accessible boorus from the passed in configuration object.
    :param config: The config object.
    :type config: ReverseSearchConfig
    """
    _all_boorus = [
        "danbooru",
        "e621",
        "iqdb",
        "paheal"
        ]

    def __init__(self, config):
        self.config = config
        self.accessible_boorus = {}

        if hasattr(self.config, "min_score"):
            if hasattr(self.config, "danbooru_username") and hasattr(self.config, "danbooru_api_key"):
                self.accessible_boorus["danbooru"] = Danbooru(self.config.danbooru_username, self.config.danbooru_api_key, self.config.min_score)
            if hasattr(self.config, "e621_username") and hasattr(self.config, "app_name") and hasattr(self.config, "version"):
                self.accessible_boorus["e621"] = E621(self.config.e621_username, self.config.app_name, self.config.version, self.config.min_score)
            self.accessible_boorus["iqdb"] = Iqdb(self.config.min_score)

        self.accessible_boorus["paheal"] = Paheal()

    async def reverse_search(self, url, callback=None, download=False):
        r"""Reverse searches all accessible boorus for ``url``.

        ``callback`` is an optional parameter that can refer to a method with a single parameter (``callback(name)``).

        After a reverse search, this method will be called with the booru name that was just searched.

        :param url: The URL to search.
        :type url: str
        :param callback: Callback function.
        :type callback: Optional[function]
        :param download: Run searches on boorus that require a file download. Defaults to False.
        :type download: Optional[bool]
        :return: A set of tags
        :rtype: Set[str]
        """
        tags = set()
        for booru in self.accessible_boorus:
            if self.accessible_boorus[booru].download_required:
                if not download:
                    continue
            tags.update(await self.search_image(booru, url))
            if callback:
                callback(booru)
        return tags


    async def search_image(self, booru, url):
        r"""Reverse search a booru for ``url``.

        :param booru: Booru to search, this must match a filename in the boorus folder.
        :type booru: str
        :param url: The URL to search.
        :type url: str
        :raises MissingAPIKeysException: Required keys in config object missing.
        :raises NotAValidBooruException: The passed in booru is not a valid booru.
        :return: A set of tags
        :rtype: Set[str]
        """
        if booru not in self._all_boorus:
            raise NotAValidBooruException("%s is not a valid booru", booru)
        if booru not in self.accessible_boorus:
            raise MissingAPIKeysException("%s is misisng one or more needed API keys. Check the documentation.")
        return set(await self.accessible_boorus[booru].search_image(url))


class ReverseSearchConfig:
    """
    Configuration object for :class:`ReverseSearch`

    All parameters are prefixed with the specific type needed for reverse searching.

    Check the relevant classes on the booru for details on the boorus.

    :param danbooru_username: Username on :class:`Danbooru`
    :type danbooru_username: str
    :param danbooru_api_key: API key on :class:`Danbooru`
    :type danbooru_api_key: 
    :param e621_username: Your :class:`E621` username
    :type e621_username: str

    :param app_name: The name of your application (required for :class:`E621`).
    :type app_name: str
    :param version: The version of your application (required for :class:`E621`).
    :type version: float

    :param min_score: Minimum search match percentage needed (required for ALL boorus except :class:`Paheal`).
    :type min_score: float
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class MissingAPIKeysException(Exception):
    """Raised if a required API key to search a booru is missing."""
    pass

class NotAValidBooruException(Exception):
    """Raised if the passed in booru does not exist."""